# encoding: utf-8
# SPDX-License-Identifier: LGPL-3.0-only

"""
Defines ``dataclass`` decorator regardless of Python version.
On versions below 3.7, attempts to use ``attr.s`` and falls back to an empty (dummy) decorator
if no implementation is found.

Similarly, it defines ``attrs`` wrapper which tries to fall back to ``dataclass`` if found.
"""

try:
	import typing as _t
	from typing import (
		Callable as _Callable,
		Union as _U,
		Protocol as _Protocol
	)

	# noinspection PyTypeHints
	_T = _t.TypeVar('T')
	# noinspection PyTypeHints
	_C = _t.TypeVar('C', bound=type)
	# noinspection PyTypeHints
	_Call = _t.TypeVar('Call', bound=callable)
except ImportError:
	pass

try:
	from dataclasses import dataclass as _dataclass_builtin
	_dataclass_import_error = None
except ImportError as _e:
	_dataclass_import_error = _e

try:
	import attr as _attr
	_attrs_import_error = None
except ImportError as _e:
	_attrs_import_error = _e

from .__py_ver import *
from .__attrs import _determine_attrs_eq_order
from .__dataclass_fallback import dataclass_fallback


def _dataclass_wrapper_py37(
	cls=None,
	init=True, repr=True, eq=True, order=False, unsafe_hash=False, frozen=False,
	match_args=True, kw_only=False, slots=False, **kwargs
):
	# type: (_C, bool, ...) -> _U[_C, _Callable[[_C], _C]]
	return _dataclass_builtin(cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash, frozen=frozen)


def _dataclass_wrapper_py310(
	cls=None,
	init=True, repr=True, eq=True, order=False, unsafe_hash=False, frozen=False,
	match_args=True, kw_only=False, slots=False, **kwargs
):
	# type: (_C, bool, ...) -> _U[_C, _Callable[[_C], _C]]
	return _dataclass_builtin(
		cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash, frozen=frozen,
		match_args=match_args, kw_only=kw_only, slots=slots,
	)


def _dataclass_with_attrs(
	cls=None,
	init=True, repr=True, eq=True, order=False, unsafe_hash=False, frozen=False,
	match_args=True, kw_only=False, slots=False, **kwargs
):
	# type: (_C, bool, ...) -> _U[_C, _Callable[[_C], _C]]
	decorator = _attr.s(
		auto_detect=False, init=init, repr=repr, eq=eq, order=order, hash=unsafe_hash,
		frozen=frozen, match_args=match_args, kw_only=kw_only, slots=slots,
	)

	if cls is None:
		return decorator
	return decorator(cls)


def dummy_decorator(func=None, **kwargs):
	# type: (_Call, ...) -> _U[_Call, _Callable[[_Call], _Call]]
	"""Decorator placeholder which does nothing."""
	def _wrap(f):
		# type: (_Call) -> _Call
		return f

	if func is None:
		return _wrap
	return _wrap(func)


def dummy_class_decorator(cls=None, **kwargs):
	# type: (_C, ...) -> _U[_C, _Callable[[_C], _C]]
	"""Same as ``dummy_decorator``, but type-hinted to be dealing with classes."""
	def _wrap(c):
		# type: (_C) -> _C
		return c

	if cls is None:
		return _wrap
	return _wrap(cls)


if _dataclass_import_error is None:
	dataclass = _dataclass_wrapper_py310 if PY310 else _dataclass_wrapper_py37
elif _attrs_import_error is None:
	dataclass = _dataclass_with_attrs
else:
	dataclass = dataclass_fallback


def _attrs_with_dataclass_py37(
	maybe_cls=None,
	these=None, repr_ns=None,

	init=True, repr=True, eq=True, order=False, hash=None, frozen=False,
	match_args=True, kw_only=False, slots=False,

	cmp=None, weakref_slot=True, str=False, auto_attribs=False,
	cache_hash=False, auto_exc=False, auto_detect=False, collect_by_mro=False, getstate_setstate=None,
	on_setattr=None, field_transformer=None,
):
	# type: (_C, ...) -> _U[_C, _Callable[[_C], _C]]
	eq, order = _determine_attrs_eq_order(cmp, eq, order, None)
	repr = repr or str

	def wrap(cls):
		# type: (_C) -> _C
		cls = _dataclass_builtin(
			cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=hash, frozen=frozen,
		)
		if str:
			cls.__str__ = cls.__repr__
		return cls

	if maybe_cls is None:
		return wrap
	return wrap(maybe_cls)


def _attrs_with_dataclass_py310(
	maybe_cls=None,
	these=None, repr_ns=None,

	init=True, repr=True, eq=True, order=False, hash=None, frozen=False,
	match_args=True, kw_only=False, slots=False,

	cmp=None, weakref_slot=True, str=False, auto_attribs=False,
	cache_hash=False, auto_exc=False, auto_detect=False, collect_by_mro=False, getstate_setstate=None,
	on_setattr=None, field_transformer=None,
):
	# type: (_C, ...) -> _U[_C, _Callable[[_C], _C]]
	eq, order = _determine_attrs_eq_order(cmp, eq, order, None)
	repr = repr or str

	def wrap(cls):
		# type: (_C) -> _C
		cls = _dataclass_builtin(
			cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=hash, frozen=frozen,
			match_args=match_args, kw_only=kw_only, slots=slots,
		)
		if str:
			cls.__str__ = cls.__repr__
		return cls

	if maybe_cls is None:
		return wrap
	return wrap(maybe_cls)


def _attrs_fallback(
	maybe_cls=None,
	these=None, repr_ns=None,

	init=True, repr=True, eq=True, order=False, hash=None, frozen=False,
	match_args=True, kw_only=False, slots=False,

	cmp=None, weakref_slot=True, str=False, auto_attribs=False,
	cache_hash=False, auto_exc=False, auto_detect=False, collect_by_mro=False, getstate_setstate=None,
	on_setattr=None, field_transformer=None,
):
	# type: (_C, ...) -> _U[_C, _Callable[[_C], _C]]
	eq, order = _determine_attrs_eq_order(cmp, eq, order, None)
	repr = repr or str

	def wrap(cls):
		# type: (_C) -> _C
		cls = dataclass_fallback(
			cls, init=init, repr=repr, eq=eq, order=order, unsafe_hash=hash, frozen=frozen,
			match_args=match_args, kw_only=kw_only, slots=slots,
		)
		if str:
			cls.__str__ = cls.__repr__
		return cls

	if maybe_cls is None:
		return wrap
	return wrap(maybe_cls)


if _attrs_import_error is None:
	attrs = _attr.s
elif _dataclass_import_error is None:
	attrs = _attrs_with_dataclass_py310 if PY310 else _attrs_with_dataclass_py37
else:
	attrs = _attrs_fallback



# encoding: utf-8
# SPDX-License-Identifier: LGPL-3.0-only

"""The last resort dataclass implementation."""

try:
	import typing as _t
	from typing import (
		Callable as _Callable,
		Union as _U,
		Protocol as _Protocol
	)

	# noinspection PyTypeHints
	_T = _t.TypeVar('T')
	# noinspection PyTypeHints
	_C = _t.TypeVar('C', bound=type)
	# noinspection PyTypeHints
	_Call = _t.TypeVar('Call', bound=callable)
except ImportError:
	pass

from .__str import str_format as _format


def _repr_attr(self, attr_nm):
	val = getattr(self, attr_nm)
	return _format('{}={}', attr_nm, repr(val))


def _repr_attr_list(self, attribs_list):
	# type: (_t.Iterable[str], _T) -> str
	arg_strings = (_repr_attr(self, nm) for nm in attribs_list)
	return _format('{}({})', self.__class__.__name__, ', '.join(arg_strings))


def _repr_func_from_attr_list(
	attribs_list,  # type: _t.Iterable[str]
):
	def _repr(self):
		return _repr_attr_list(self, attribs_list)
	return _repr


def _repr_from_slots(cls):
	# type: (_C) -> _C
	setattr(cls, '__repr__', _repr_func_from_attr_list(cls.__slots__))
	return cls


@_repr_from_slots
class _DataclassParams(object):
	__slots__ = (
		'init',
		'repr', 'eq', 'order', 'unsafe_hash', 'frozen',
		'match_args', 'kw_only', 'slots',
	)

	def __init__(
		self,
		init=True, repr=True, eq=True, order=False, unsafe_hash=False, frozen=False,
		match_args=True, kw_only=False, slots=False, **kwargs
	):
		self.init = bool(init)
		self.repr = bool(repr)
		self.eq = bool(eq)
		self.order = bool(order)
		self.unsafe_hash = bool(unsafe_hash)
		self.frozen = bool(frozen)

		self.match_args = bool(match_args)
		self.kw_only = bool(kw_only)
		self.slots = bool(slots)


_DC_PARAMS = '__dataclass_params__'
_DC_FIELDS_TUPLE = '__dataclass_fields_tuple__'  # NON-default! Stored "just in case".


def _class_field_items(cls):
	# type: (_C) -> _t.Tuple[_t.Tuple[str, _t.Any], ...]
	field_names = (x for x in dir(cls) if not x.startswith('_'))
	field_items = (
		(k, getattr(cls, k, None)) for k in field_names
	)
	return tuple(
		(k, v) for k, v in field_items
		if not callable(v)
	)


def _process_class(
	cls,  # type: _C
	**dc_kwargs
):
	"""The main implementation of dataclass."""
	dc_params = _DataclassParams(**dc_kwargs)
	setattr(cls, _DC_PARAMS, dc_params)

	field_names = tuple(k for k, v in _class_field_items(cls))
	field_names_set = set(field_names)

	setattr(cls, _DC_FIELDS_TUPLE, field_names)

	def _init(self, **init_kwargs):
		for k, v in init_kwargs.items():
			if k not in field_names_set:
				raise TypeError(_format(
					"No such field in dataclass {}: {}", self.__class__.__name__, repr(k)
				))
			setattr(self, k, v)

	def as_tuple(self):
		return tuple(
			getattr(self, x, None) for x in field_names
		)

	def _eq(self, other):
		if other.__class__ is not self.__class__:
			return NotImplemented
		return as_tuple(self) == as_tuple(other)

	for do_attach, method_name, func in (
		(dc_params.init, '__init__', _init),
		(dc_params.repr, '__repr__', _repr_func_from_attr_list(field_names)),
		(dc_params.eq, '__eq__', _eq),
	):
		if do_attach:
			setattr(cls, method_name, func)

	return cls


def dataclass_fallback(
	cls=None,
	init=True, repr=True, eq=True, order=False, unsafe_hash=False, frozen=False,
	match_args=True, kw_only=False, slots=False, **kwargs
):
	# type: (_C, ...) -> _U[_C, _Callable[[_C], _C]]
	"""
	A simplest possible re-implementation as fallback when neither built-in ``dataclass`` nor ``attr.s`` are available.

	It should be used only as the very last resort.
	It doesn't provide any of advanced features and currently implements only:
		* init
		* repr
		* eq

	All the other arguments do nothing and are there just for signature consistency.
	"""
	def wrap(cls):
		return _process_class(
			cls,
			init=init, repr=repr, eq=eq, order=order, unsafe_hash=unsafe_hash, frozen=frozen,
			match_args=match_args, kw_only=kw_only, slots=slots, **kwargs
		)

	if cls is None:
		return wrap
	return wrap(cls)



# encoding: utf-8
# SPDX-License-Identifier: LGPL-3.0-only

"""
"""

try:
	import typing as _t
except ImportError:
	pass

from .__py_ver import *


# noinspection PyBroadException
try:
	# noinspection PyShadowingBuiltins
	unicode = unicode
except Exception:
	# noinspection PyShadowingBuiltins
	unicode = str


def to_least_str(val):
	# type: (...) -> _t.AnyStr
	"""
	Python 2:
		* Try to convert to ``str()``. If fails, convert to ``unicode()``.
		*
			For custom classes inherited from either of them, try to preserve it
			(`unicode` subclass might be converted to regular `str` if possible).

	Python 3:
		Just an alias for ``str()``.
	"""
	if isinstance(val, str):
		return val

	# noinspection PyBroadException
	try:
		return str(val)
	except Exception:
		if isinstance(val, unicode):
			return val
		return unicode(val)


def str_format(format_pattern, *args, **kwargs):
	# type: (_t.AnyStr, _t.Any, _t.Any) -> _t.AnyStr
	"""
	Python 2:
		Unicode-safe string format. Just performs ``unicode.format()`` if ``str.format()`` fails.

	Python 3:
		Just an alias for ``str.format``.
	"""
	# noinspection PyBroadException
	try:
		return format_pattern.format(*args, **kwargs)
	except Exception:
		return unicode(format_pattern).format(*args, **kwargs)


if PY3:
	to_least_str = str
	str_format = str.format



#!/usr/bin/env python
# encoding: utf-8

"""
https://packaging.python.org/en/latest/guides/distributing-packages-using-setuptools/#setup-args
"""

from os import path
import sys

from setuptools import setup

import darlog_pypi

HERE = path.abspath(path.dirname(__file__))
SRC_DIR = "src"

sys.path.append(path.join(HERE, SRC_DIR))
from darlog import py23 as the_module


NAME = the_module.__pypi_package__
SHORT_DESC = the_module.__doc__.strip()

# We use this to cross-link PyPi <-> GitHub pages (header in readme text is updated during build):
LONG_DESC = darlog_pypi.ReadmeUpdater.from_rel_path(HERE, "README.md", NAME).update_for_github().text_for_pypi()


setup(
	name=NAME,
	version=the_module.__version__,
	description=SHORT_DESC,
	long_description=LONG_DESC,
	long_description_content_type="text/markdown",
		# https://packaging.python.org/en/latest/specifications/core-metadata/#description-content-type-optional

	url=the_module.__url__,
	author=the_module.__author__,
	maintainer=the_module.__author__,
	license=the_module.__license__,
)
